﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

using System.IO.Ports;

namespace ServoAdj
{
    // COMポート
    public partial class ComPort : UserControl
    {
        /// <summary>
        /// COMポート受信時イベント
        /// </summary>
        public event ComPortEventHandler onReceived;
        /// <summary>
        /// COMポート接続時イベント
        /// </summary>
        public event ComPortEventHandler onOpened;
        /// <summary>
        /// COMポート切断時イベント
        /// </summary>
        public event ComPortEventHandler onClosed;

        SerialPort serialPort;
        string portName = "COM1";
        int baudRate = 9600;

        // 受信状態
        int rxState;
        // 電文開始待ち状態
        const int STATE_READY = 0;
        // 電文受信中状態
        const int STATE_RECEIVING = 1;

        // 受信バッファ
        int rxPtr;
        char[] rxBuff;

        // 開始文字・終了文字　（可変にするためconstとはしない）
        char STX_CODE = (char)0x02;
        char ETX_CODE = (char)0x03;

        public ComPort()
        {
            InitializeComponent();

            updateList();

            rxState = STATE_READY;
            rxPtr = 0;
            rxBuff = new char[5000];

            serialPort = new SerialPort();
            serialPort.DataReceived += new SerialDataReceivedEventHandler(serialPort_DataReceived);
        }

        /// <summary>
        /// このコントロールの表示名
        /// </summary>
        [Browsable(true)]
        public string NameText
        {
            get
            {
                return m_text;
            }
            set
            {
                m_text = value;
                groupBox.Text = value;
            }
        }
        private string m_text;

        /// <summary>
        /// デフォルトのCOMポート名
        /// </summary>
        [Browsable(true)]
        public string DefaultPortName
        {
            get
            {
                return m_defaultPortName;
            }
            set
            {
                m_defaultPortName = value;
            }
        }
        private string m_defaultPortName;

        /// <summary>
        /// COMポートは開いているか
        /// </summary>
        public bool isOpen
        {
            get
            {
                return serialPort.IsOpen;
            }
        }

        /// <summary>
        /// 開始文字・終了文字を設定する
        /// </summary>
        /// <param name="stx">開始文字</param>
        /// <param name="etx">終了文字</param>
        /// <returns>成否</returns>
        public bool setStxEtx(char stx, char etx)
        {
            STX_CODE = stx;
            ETX_CODE = etx;
            return true;
        }
        
        /// <summary>
        /// ボーレートを設定する
        /// </summary>
        /// <param name="baud">ボーレート</param>
        /// <returns>成否</returns>
        public bool setBaudRate(int baud)
        {
            baudRate = baud;

            return true;
        }

        /// <summary>
        /// 開く
        /// </summary>
        /// <returns>成否</returns>
        public bool open()
        {
            if (serialPort.IsOpen) return true;

            // COMポートを開く
            try
            {
                // COMポートを開く
                serialPort.PortName = portName;
                serialPort.BaudRate = baudRate;
                serialPort.Parity = System.IO.Ports.Parity.None;
                serialPort.DataBits = 8;
                serialPort.StopBits = System.IO.Ports.StopBits.One;
                serialPort.Handshake = System.IO.Ports.Handshake.None;
                serialPort.Encoding = Encoding.UTF8;
                serialPort.NewLine = "\r";
                serialPort.Open();

                // イベントハンドラを呼ぶ
                ComPortEventArgs args = new ComPortEventArgs();
                onOpened(this, args);
            }
            catch
            {
                serialPort.Close();
                return false;
            }
            return true;
        }
        /// <summary>
        /// 開く
        /// </summary>
        /// <param name="port">ポート名</param>
        /// <returns>成否</returns>
        public bool open(string port)
        {
            portName = port;
            return open();
        }
        /// <summary>
        /// 開く
        /// </summary>
        /// <param name="port">ポート名</param>
        /// <param name="baud">ボーレート</param>
        /// <returns>成否</returns>
        public bool open(string port, int baud)
        {
            portName = port;
            baudRate = baud;
            return open();
        }

        /// <summary>
        /// 閉じる
        /// </summary>
        /// <returns>成否</returns>
        public bool close()
        {
            if (!serialPort.IsOpen) return true;

            // COMポートを閉じる
            try
            {
                serialPort.Close();

                // イベントハンドラを呼ぶ
                ComPortEventArgs args = new ComPortEventArgs();
                onClosed(this, args);
            }
            catch
            {
                ;
            }
            return true;
        }

        /// <summary>
        /// 文字列を送信する
        /// </summary>
        /// <param name="data">送信する文字列</param>
        /// <returns></returns>
        public bool send(string data)
        {
            if (serialPort.IsOpen)
            {
                try
                {
                    this.serialPort.Write(data);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message, "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                return true;
            }
            else
            {
                return false;
            }
        }

        /// <summary>
        /// メッセージを送信する（開始文字と終了文字が付加されて送信される）
        /// </summary>
        /// <param name="message"></param>
        /// <returns></returns>
        public bool sendMessage(string message)
        {
            string data = STX_CODE.ToString() + message + ETX_CODE.ToString();

            return send(data);
        }

        // 受信ハンドラ
        private void serialPort_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            // 受信文字列の取得
            string receivedString;
            try
            {
                receivedString = this.serialPort.ReadExisting();
                byte[] data = System.Text.Encoding.ASCII.GetBytes(receivedString);

                char c;
                for (int i = 0; i < data.Length; i++)
                {
                    //Serial.println("RECV ");
                    c = (char)data[i];

                    switch (rxState)
                    {
                        /* 電文開始待ち状態 */
                        case STATE_READY:
                            /* 電文開始コードが来たら電文受信中状態へ */
                            if (c == STX_CODE)
                            {
                                //Serial.println("STX ");
                                rxState = STATE_RECEIVING;
                                rxPtr = 0;
                            }
                            break;
                        /* 電文受信中状態 */
                        case STATE_RECEIVING:
                            /* もしも電文開始コードが来たら受信中のデータを破棄 */
                            if (c == STX_CODE)
                            {
                                //Serial.println("STX ");
                                rxPtr = 0;
                            }
                            /* 電文終了コードが来たら、受信した電文のコマンドを実行 */
                            else if (c == ETX_CODE)
                            {
                                //Serial.println("ETX ");
                                rxBuff[rxPtr] = '\0';

                                // イベントハンドラを呼ぶ
                                ComPortEventArgs args = new ComPortEventArgs();
                                args.data = rxBuff;
                                onReceived(this, args);

                                rxState = STATE_READY;
                            }
                            /* 1文字受信 */
                            else
                            {
                                rxBuff[rxPtr] = c;
                                rxPtr++;
                                if (rxPtr >= 5000)
                                {
                                    rxState = STATE_READY;
                                }
                            }
                            break;
                        default:
                            rxState = STATE_READY;
                            break;
                    }

                }
            }
            catch (Exception ex)
            {
                receivedString = ex.Message;
            }
        }

        /// <summary>
        /// COMポートのリストを更新する
        /// </summary>
        public void updateList()
        {
            // コンボボックスのクリア
            comboPortList.Items.Clear();
            // すべてのシリアル・ポート名を取得する
            string[] ports = SerialPort.GetPortNames();
            // 取得したシリアル・ポート名をリストに加える
            bool matched = false;
            foreach (string port in ports)
            {
                comboPortList.Items.Add(port);
                if (port.Equals(m_defaultPortName))
                {
                    comboPortList.SelectedItem = port;
                    matched = true;
                }
            }
            if (!matched && (comboPortList.Items.Count > 0)) comboPortList.SelectedIndex = 0;
        }

        // 更新ボタン
        private void buttonUpdate_Click(object sender, EventArgs e)
        {
            updateList();
        }

        // 開く/閉じるボタン
        private void buttonOpen_Click(object sender, EventArgs e)
        {
            // 開いていたら閉じる
            if (serialPort.IsOpen)
            {
                close();
                buttonOpen.Text = "開く";
            }
            // 閉じていたら開く
            else
            {
                if (open(comboPortList.Text))
                {
                    buttonOpen.Text = "閉じる";
                }
            }
        }

    }

    /// <summary>
    /// COMポートイベントの引数
    /// </summary>
    public class ComPortEventArgs : EventArgs
    {
        /// <summary>
        /// 受信文字列
        /// </summary>
        public char[] data;
    }
    public delegate void ComPortEventHandler(object sender, ComPortEventArgs e);
}
